/*-------------------------------------------------------------------------------------------------------------*
   OS/2 Dynamic Job Properties Example Program

   The purpose of this program is to demonstrate OS/2's ability to change
   job properties on a per page basis without the need to display the job
   properties dialog and start a separate print job for each page.  The following
   code implements a simple frame window with a File menu that has two
   options: Print Dynamic Job Properties Example and Close.  When Print...
   is selected, the application spawns a thread that sends three pages to the
   default printer.  Page 1 should be landscape, page 2 should be portrait,
   and page 3 should go back to landscape.  This should all happen without
   affecting the printer's original job properties.

   DISCLAIMER OF WARRANTIES.  The following code is
   sample code created by J&J Computer Consulting. This sample code
   is provided to you solely for  the purpose of assisting you in the
   development of your applications.  The code is provided "AS IS", without
   warranty of any kind.  J&J Computer Consulting shall not be liable for any
   damages arising out of your use of the sample code, even if they have been
   advised of the possibility of   such damages.

   Copyright 1998 J&J Computer Consulting

*------------------------------------------------------------------------------------------------------------------*/

#define INCL_PM
#define INCL_BASE
#define INCL_DOS
#define INCL_DOSPROCESS
#define INCL_DEV
#define INCL_DEVDJP
#define INCL_SPL
#define INCL_SPLDOSPRINT
#define INCL_SPLERRORS
#include <os2.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include "os2print.h"

int main()

{

 /* Typical OS/2 PM main function implemented in C.  Create the frame window and start
     accepting messages.  Since this example is not about PM programming, this section
     contains no further comments. */

   HAB               hab;
   HMQ              hmq;
   ULONG         flCreateFlags = FCF_TITLEBAR      |
                                                      FCF_SYSMENU       |
                                                      FCF_SIZEBORDER    |
                                                      FCF_MINMAX        |
                                                      FCF_SHELLPOSITION |
                                                      FCF_TASKLIST      |
                                                      FCF_MENU;

   QMSG                                      qmsg;
   HWND                                      hwndClient, hwndFrame;

   hab = WinInitialize (0);
   hmq = WinCreateMsgQueue (hab, 0);

   if (!WinRegisterClass (hab,
                                             WIN_CLASS_NAME,
                                             fnClientWndProc,
                                             CS_SIZEREDRAW,
                                             0 )) {
      exit(EXIT_FAILURE);
   }

   if (!(hwndFrame = WinCreateStdWindow (HWND_DESKTOP,
                                                                                   WS_VISIBLE,
                                                                                    &flCreateFlags,
                                                                                    WIN_CLASS_NAME,
                                                                                    "OS/2 Dynamic Job Properties Example",
                                                                                    0,
                                                                                    NULLHANDLE,
                                                                                    IDW_FRAME_WINDOW,
                                                                                    &hwndClient)))    {
      int rc = WinGetLastError( hab );
      exit(EXIT_FAILURE);
   }

   while (WinGetMsg (hab, &qmsg, 0, 0, 0)) {
      WinDispatchMsg (hab, &qmsg);
   }

   if (WinIsWindow (hab, hwndFrame)) {
      WinDestroyWindow (hwndFrame);
   }

   WinDestroyMsgQueue (hmq);
   WinTerminate (hab);

   return 0;
}


MRESULT EXPENTRY fnClientWndProc (HWND hwnd, ULONG msg, MPARAM   mp1, MPARAM   mp2) {

 MRESULT     mr = (MRESULT) FALSE;

   switch (msg) {
        case WM_CREATE: {
            break;
        }

        case WM_PAINT: {
            HPS hps;
            RECTL rectClientWindow;
            hps = WinBeginPaint (hwnd,
                                 NULLHANDLE,
                                 &rectClientWindow);
            WinQueryWindowRect (hwnd, &rectClientWindow);
            WinFillRect (hps,
                         &rectClientWindow,
                         SYSCLR_WINDOW);
            static char title[] = "J&J Computer Consulting - http://www.ceri-net.com/jjcomp";
            POINTL point;
            point.x = 0;
            point.y = 5;
            GpiCharStringAt( hps, &point, strlen( title ), title );
            WinEndPaint (hps);
            break;
        }

        case WM_COMMAND:

           switch (SHORT1FROMMP(mp1)) {
               case IDM_FILE_PRINT:   {
                  /* On File Print message, spawn the print thread.  */
                  _beginthread( printSample, NULL, 64000, (void*) NULL );
                  break;
               }
               case IDM_FILE_CLOSE:
                  WinPostMsg( hwnd, WM_CLOSE, 0, 0 );
                  break;
           };
           break;

        case WM_CLOSE: {
           WinPostMsg (hwnd, WM_QUIT, 0, 0);
           break;
        }
        default:
           mr = WinDefWindowProc (hwnd, msg, mp1, mp2);
           break;
   };

   return mr;
}

/* printSample: Queries the OS/2 initialization file for the default printer and prints 3 pages
    to it using dynamic job properties to alternate the page orientation. */

void _Optlink printSample( void* pvoidargs ) {

   char printerData[ 256 ];
   DRIVDATA drivData = { sizeof( DRIVDATA ) };
   char defaultQueue[ 256 ];
   char* tmp;

   HAB hab = WinInitialize(0);

   //Get the default queue name from the initialization file
   PrfQueryProfileString( HINI_PROFILE,
                                             "PM_SPOOLER",
                                             "QUEUE",
                                             NULL,
                                             defaultQueue,
                                             256 );

   //Terminate the queue name at the semi-colon
   tmp = strchr( defaultQueue, ';' );
   if( tmp ) {
      tmp[ 0 ] = '\0';
   }

   unsigned long bytesNeeded;

   //Call SplQueryQueue without buffers to determine the size needed for the PRQINFO6 structure
   //it will return on the next try.
   unsigned long rc = SplQueryQueue( NULL,
                                             defaultQueue,
                                             6L, //Information level 6.  Nice to use 6 instead of 3 so you get network info.
                                             (PVOID)NULL,
                                             0L,
                                             &bytesNeeded );

   //Make sure that the error returned was what was expected
   if (rc != NERR_BufTooSmall && rc != ERROR_MORE_DATA ) {
      rc = WinGetLastError( hab );
      WinTerminate( hab );
      _endthread();
   }

   //Allocate memory for the PRQINFO6 structure.  Since it's not a fixed size structure, allocate
   //it as bytes.
   char* pBuf = new char[ bytesNeeded ];
   unsigned long bytesBuf = bytesNeeded;

   //Go ahead and fill the structure on this call
   rc = SplQueryQueue(  NULL,
                                               defaultQueue,
                                               6L,
                                               pBuf,
                                               bytesBuf,
                                               &bytesNeeded) ;

   PPRQINFO6 prq6=(PPRQINFO6)pBuf; //Assign the buffer to a PRQINFO6 pointer

   //Create a DEVOPENSTRUC for use in the call to DevOpenDC
   DEVOPENSTRUC dev;
   dev.pszLogAddress = defaultQueue;
   char *driver = new char[ strlen( prq6->pszDriverName ) + 1 ];
   strcpy( driver, prq6->pszDriverName );
   tmp = strchr( driver, '.' ); //The driver name used in this structure cannot contain the string after the '.'
   if( tmp ) {
      tmp[ 0 ] = '\0';
   }
   dev.pszDriverName = driver;
   dev.pdriv = prq6->pDriverData; //These are the default printers original job properties
   dev.pszDataType = "PM_Q_STD";
   dev.pszComment = "OS/2 Dynamic Job Properties Example";
   dev.pszQueueProcName = NULL;
   dev.pszQueueProcParams = "";
   dev.pszSpoolerParams = "";

   //Call DevOpenDC to get a handle to a device context for the default printer.
   HDC hdc = DevOpenDC( hab,
                                                      OD_QUEUED,
                                                      "*",
                                                      4L,
                                                      (PDEVOPENDATA)&dev,
                                                      0L );

   if( hdc == DEV_ERROR ) {
      rc = WinGetLastError( hab );
      WinTerminate( hab );
      _endthread();
   }

   SIZEL size = { 0, 0 };

   //Create a presentation space to use for writing to the printer.  Note that PU_TWIPS
   //is used for the units.  It's very easy to work with PU_TWIPS for text output, but
   //dynamic job properties seems to have thrown a small wrench into the works.  On
   //the first call to DevEscape in which the job properties are changed, the units switch
   //to PU_PELS.
   HPS hps = GpiCreatePS( hab,
                                          hdc,
                                          &size,
                                          PU_TWIPS | GPIT_NORMAL | GPIF_DEFAULT | GPIA_ASSOC );

   //FYI - Notice a lot of cleanup work left out for error conditions.  Don't use this code
   //in a real program without fixing these leaks and problems.
   if( hps == 0 ) {
      rc = WinGetLastError( hab );
      WinTerminate( hab );
      _endthread();
   }
   char jobName[] = "OS/2 Dynamic Job Properties Example";

   //First look at dynamic job properties.  This is the DJP_ITEM structure used
   //to ask the driver to change the properties for you, or the driver can also return
   //properties in this structure so your application can determine the current
   //settings.  For this application, create DJP_ITEM list for changing the page orientation
   DJP_ITEM djp[ 2 ]; //one for landscape, one to terminate

   djp[ 0 ].cb = sizeof( DJP_ITEM );  //Standard OS/2 sizeof in structure's first element
   djp[ 0 ].ulProperty = DJP_SJ_ORIENTATION; //Property id for page orientation.  Find others in pmdjp.h
   //Not sure about the lType member.  May be to tell the driver to change for the current job only or
   //for all jobs.  Current would seem to be what most any application would use.
   djp[ 0 ].lType = DJP_CURRENT;
   djp[ 0 ].ulNumReturned = 0; //Used when querying job properties, not 100% sure of its value.
   djp[ 0 ].ulValue = DJP_ORI_LANDSCAPE; //Value for a landscape page orientation

   djp[ 1 ].cb = sizeof( DJP_ITEM );
   djp[ 1 ].ulProperty = DJP_NONE; //Driver reads these as a list.  DJP_NONE property terminates the list.
   djp[ 1 ].lType = DJP_CURRENT;
   djp[ 1 ].ulNumReturned = 0;
   djp[ 1 ].ulValue = 0;

   long jobpropSize = (long)bytesNeeded;

   //Unnecessary call.  Used to demonstrate that you can get job properties from the driver into
   //one of these structures.  Unfortunately, some of the more informative properties have complex
   //strucutres that are not completely clear to me yet how they fit into DJP_ITEM.ulValue.

   //Note that all this call really does is look at prq6->pDriverData and interpret the traditionally
   //vendor dependent string into DJP_ITEM structures.  The list of DJP_ITEM structures you pass
   //in tells the driver which properties to return.
   rc = DevEscape( hdc,
                                     DEVESC_QUERYJOBPROPERTIES,
                                     2*sizeof( DJP_ITEM ),
                                     (PBYTE)djp,
                                     &jobpropSize,
                                     (PBYTE)prq6->pDriverData );

   if( rc == DEVESC_ERROR ) {
      rc = WinGetLastError( hab );
   }

   djp[ 0 ].ulValue = DJP_ORI_LANDSCAPE; //Query call may have switched this back to portrait.

   //While the name of the escape implies it does something, don't be fooled.  All it does it take
   //what is in the list of DJP_ITEM structures and modify prq6->pDriverData appropriately.
   rc = DevEscape( hdc,
                                    DEVESC_SETJOBPROPERTIES,
                                    2*sizeof( DJP_ITEM ),
                                    (PBYTE)djp,
                                    //It's possible this call could change the size of pDriverData.  If so, you could
                                    //get an error for the buffer not being large enough on some drivers.
                                    &jobpropSize,
                                    (PBYTE)prq6->pDriverData );

   if( rc == DEVESC_ERROR ) {
      rc = WinGetLastError( hab );
   }

   //This call is where something actually happens.  Taking the job properties that were configured
   //above, call DEVESC_STARTDOC_WPROP (with properties) and pass the job properties
   //as parameters.  The driver will change the printer's properties accordingly.
   if( DevEscape( hdc,
                                 DEVESC_STARTDOC_WPROP,
                                 strlen( jobName ),
                                 jobName,
                                 &jobpropSize,
                                 (PBYTE)prq6->pDriverData ) != DEVESC_ERROR ) {

         //Create a font
         FATTRS fattrs;
         fattrs.usRecordLength = sizeof( fattrs );
         fattrs.fsSelection = 0;
         fattrs.lMatch = 0;
         fattrs.idRegistry = 0;
         fattrs.usCodePage = GpiQueryCp( hps );
         fattrs.lMaxBaselineExt = 0;
         fattrs.lAveCharWidth = 0;
         fattrs.fsType = 0;
         fattrs.fsFontUse = FATTR_FONTUSE_OUTLINE | FATTR_FONTUSE_TRANSFORMABLE;
         strcpy( fattrs.szFacename, "Helv" );
         GpiCreateLogFont( hps, NULL, 1L, &fattrs );
         GpiSetCharSet( hps, 1L );
         POINTL point;
         LONG outlen = 0;
         PBYTE outbuf = NULL;

         char page1[] = "OS/2 Dynamic Job Properties Example: Page 1 - This Should be Landscape";
         char page2[] = "OS/2 Dynamic Job Properties Example: Page 2 - This Should be Portrait";
         char page3[] = "OS/2 Dynamic Job Properties Example: Page 3 - This Should be Landscape";

         //Center the text
         if( !GpiSetTextAlignment( hps, TA_CENTER, TA_TOP ) ) {
            rc = WinGetLastError( hab );
         }
         SIZEL pageSize;
         rc = GpiQueryPS( hps, &pageSize );
         point.x = pageSize.cx/2;
         point.y = pageSize.cy/2;

         //Print the string
         GpiCharStringAt( hps, &point, strlen( page1 ), page1 );

         djp[ 0 ].ulValue = DJP_ORI_PORTRAIT; //set the orientation in the DJP_ITEM structures to portrait

         //Ask the driver to set prq6->pDriverData (the job properties) to portrait.  Page orientation does
         //not actually change with this call.
         rc = DevEscape( hdc,
                                          DEVESC_SETJOBPROPERTIES,
                                          2*sizeof( DJP_ITEM ),
                                          (PBYTE)djp,
                                          &jobpropSize,
                                          (PBYTE)prq6->pDriverData );

         if( rc == DEVESC_ERROR ) {
            rc = WinGetLastError( hab );
         }

         //The call to DEVESC_NEWFRAME_WPROP creates a new page and changes the
         //job properties based on the properties passed as parameters, much like start doc
         //did.
         rc = DevEscape( hdc,
                                           DEVESC_NEWFRAME_WPROP,
                                           0L,
                                           NULL,
                                           &jobpropSize,
                                           (PBYTE)prq6->pDriverData );

         if( rc == DEVESC_ERROR ) {
            rc = WinGetLastError( hab );
         }

         rc = GpiQueryPS( hps, &pageSize );
         point.x = pageSize.cx/2;
         point.y = pageSize.cy/2;
         GpiCharStringAt( hps, &point, strlen( page2 ), page2 );

         djp[ 0 ].ulValue = DJP_ORI_LANDSCAPE;  //back to landscape

         rc = DevEscape( hdc,
                                          DEVESC_SETJOBPROPERTIES,
                                          2*sizeof( DJP_ITEM ),
                                          (PBYTE)djp,
                                          &jobpropSize,
                                          (PBYTE)prq6->pDriverData );

         if( rc == DEVESC_ERROR ) {
            rc = WinGetLastError( hab );
         }

         rc = DevEscape( hdc,
                                           DEVESC_NEWFRAME_WPROP,
                                           0L,
                                           NULL,
                                           &jobpropSize,
                                           (PBYTE)prq6->pDriverData );

         if( rc == DEVESC_ERROR ) {
            rc = WinGetLastError( hab );
         }

         rc = GpiQueryPS( hps, &pageSize );
         point.x = pageSize.cx/2;
         point.y = pageSize.cy/2;
         GpiCharStringAt( hps, &point, strlen( page3 ), page3 );

         //End the document
         rc = DevEscape( hdc, DEVESC_ENDDOC, 0L, NULL, NULL, NULL );

         GpiSetCharSet( hps, LCID_DEFAULT ); //Remove the font that was created
         GpiDeleteSetId( hps, 1L ); //Delete the created font
   }
   else {
      rc = WinGetLastError( hab );
   }
   GpiDestroyPS( hps ); //Destroy the presentation space
   DevCloseDC( hdc ); //Close the device context
   WinTerminate( hab );  //Terminate and end the thread
   _endthread();
}

